import { EventEmitter } from 'events';
import * as Promise from 'bluebird';
import * as zk from 'node-zookeeper-client';
import { Locator } from 'locators';
/**
 * Error thrown by locking action when blocking wait for lock reaches a timeout period
 */
export declare class ZookeeperLockTimeoutError extends Error {
    lockPath: string;
    timeout?: number;
    constructor(message: string, path: string, timeout?: number);
}
/**
 * Error thrown by locking action when config.failImmediate == true when a lock is already locked
 */
export declare class ZookeeperLockAlreadyLockedError extends Error {
    lockPath: string;
    constructor(message: string, path: string);
}
export declare class Configuration {
    serverLocator?: Locator;
    pathPrefix?: string;
    sessionTimeout?: number;
    spinDelay?: number;
    retries?: number;
    failImmediate?: boolean;
    maxConcurrentHolders?: number;
}
export declare class ZookeeperLock extends EventEmitter {
    path: string;
    key: string;
    private config;
    client: zk.Client;
    private connected;
    private timedOut;
    private locked;
    private disconnecting;
    private retryCount;
    private static config;
    static Signals: {
        LOST: string;
        TIMEOUT: string;
    };
    /**
     * create a new zk lock
     * @param config
     */
    constructor(config: Configuration);
    /**
     * create a zookeeper client which powers the lock, done when creating a new lock
     * or zk connection expires
     * @returns {Promise<any>}
     */
    private createClient();
    /**
     * connect underlying zookeeper client, with optional delay
     * @param [delay=0]
     * @returns {Promise<any>}
     */
    connect: (delay?: number) => Promise<any>;
    private connectHelper;
    /**
     * disconnect zookeeper client, and remove all event listeners from it
     * @returns {Promise<any>}
     */
    disconnect: () => Promise<any>;
    private disconnectHelper;
    /**
     * destroy the lock, disconnect and remove all listeners from the 'signal' event emitter
     * @returns {Promise<any>}
     */
    destroy: () => Promise<boolean>;
    /**
     * internal method to reconnect, wired up to disconnect event of zk client
     * @returns {Promise<any>}
     */
    private reconnect;
    /**
     * unlock a lock, removing the key from zookeeper, and disconnecting
     * the zk client and all event listeners. By default this also destroys
     * the lock and removes event listeners on the locks 'signals' event
     * @param [destroy=true] - remove listeners from lock in addition
     * to disconnecting zk client on completion, defaults to true
     * @returns {Promise<any>}
     */
    unlock: (destroy?: boolean) => Promise<any>;
    /**
     * wait for a lock to become free for a given key and acquire it, with an optional
     * timeout upon which the lock will fail. if not currently connected to zookeeper,
     * this will connect, and on timeout, the lock will disconnect from zookeeper
     * @param key
     * @param [timeout]
     * @returns {Promise<any>}
     */
    lock: (key: string, timeout?: number) => Promise<any>;
    private lockHelper;
    /**
     * make the zk node that will hold the locks if it doens't already exist
     * @param path
     * @returns {Promise<any>}
     */
    private makeLockDir;
    /**
     * create a lock as a ephemeral sequential child node of the supplied path, prefixed with 'lock-',
     * to state intent to acquire a lock
     * @param path
     * @returns {Promise<any>}
     */
    private initLock;
    /**
     * loop until lock is available or timeout occurs
     * @param path
     * @returns {Promise<any>}
     */
    private waitForLock;
    private checkBail;
    /**
     * helper method that does the grunt of the work of waiting for the lock. This method does 2 things, first
     * reads the lock path to compare the locks key to the other keys that are children of the path. if this locks
     * sequence number is the lowest, the lock has been aquired. If not, this method reactively responds to
     * children changed events from the zk-client for the path we want to aqcuire the lock for, and recurses to
     * repeat this process until the sequence is the lowest
     * @param resolve
     * @param reject
     * @param path
     */
    private waitForLockHelper;
    /**
     * method to filter zk node children to contain only those that are prefixed with 'lock-',
     * which are assumed to be created by this library
     * @param children
     * @returns {string[]|T[]}
     */
    private filterLocks;
    /**
     * check if a lock exists, connecting to zk client if not connected
     * @param key
     * @returns {Promise<boolean>}
     */
    checkLocked: (key: string) => Promise<boolean>;
    private checkedLockedHelper;
    /**
     * set static config to use by static helper methods
     * @param config
     */
    static initialize: (config: any) => void;
    /**
     * create a new lock using the static stored config
     * @returns {ZookeeperLock}
     */
    static lockFactory: () => ZookeeperLock;
    /**
     * create a new lock and lock it using the static stored config, with optional timeout
     * @param key
     * @param timeout
     * @returns {Promise<ZookeeperLock>}
     */
    static lock: (key: string, timeout?: number) => Promise<ZookeeperLock>;
    /**
     * check if a lock exists for a path using the static config
     * @param key
     * @returns {Promise<boolean>}
     */
    static checkLock: (key: string) => Promise<boolean>;
    /**
     * get the numeric part of the lock key
     * @param path
     * @returns {number}
     */
    private static getSequenceNumber;
}
