"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var events_1 = require('events');
var Promise = require('bluebird');
var zk = require('node-zookeeper-client');
var util = require('util');
var debuglog = util.debuglog('zk-lock');
/**
 * Error thrown by locking action when blocking wait for lock reaches a timeout period
 */
var ZookeeperLockTimeoutError = (function (_super) {
    __extends(ZookeeperLockTimeoutError, _super);
    function ZookeeperLockTimeoutError(message, path, timeout) {
        _super.call(this, message);
        this.message = message;
        this.lockPath = path;
        this.timeout = timeout;
    }
    return ZookeeperLockTimeoutError;
}(Error));
exports.ZookeeperLockTimeoutError = ZookeeperLockTimeoutError;
/**
 * Error thrown by locking action when config.failImmediate == true when a lock is already locked
 */
var ZookeeperLockAlreadyLockedError = (function (_super) {
    __extends(ZookeeperLockAlreadyLockedError, _super);
    function ZookeeperLockAlreadyLockedError(message, path) {
        _super.call(this, message);
        this.message = message;
        this.lockPath = path;
    }
    return ZookeeperLockAlreadyLockedError;
}(Error));
exports.ZookeeperLockAlreadyLockedError = ZookeeperLockAlreadyLockedError;
var Configuration = (function () {
    function Configuration() {
    }
    return Configuration;
}());
exports.Configuration = Configuration;
var ZookeeperLock = (function (_super) {
    __extends(ZookeeperLock, _super);
    /**
     * create a new zk lock
     * @param config
     */
    function ZookeeperLock(config) {
        var _this = this;
        _super.call(this);
        this.config = null;
        this.client = null;
        this.connected = false;
        this.timedOut = false;
        this.locked = false;
        this.disconnecting = false;
        this.retryCount = 0;
        /**
         * connect underlying zookeeper client, with optional delay
         * @param [delay=0]
         * @returns {Promise<any>}
         */
        this.connect = function (delay) {
            if (delay === void 0) { delay = 0; }
            debuglog('connecting...');
            _this.disconnecting = false;
            return Promise.delay(delay).then(function () {
                if (_this.connected) {
                    debuglog('already connnected');
                    return true;
                }
                if (_this.client === null) {
                    debuglog('client null, creating...');
                    return _this.createClient().then(function () {
                        return _this.connectHelper();
                    });
                }
                return _this.connectHelper();
            });
        };
        this.connectHelper = function () {
            return new Promise(function (resolve, reject) {
                _this.client.once('connected', function () {
                    debuglog('connected');
                    _this.connected = true;
                    resolve(true);
                });
                _this.client.connect();
            });
        };
        /**
         * disconnect zookeeper client, and remove all event listeners from it
         * @returns {Promise<any>}
         */
        this.disconnect = function () {
            if (_this.key && _this.path) {
                debuglog(_this.path + "/" + _this.key + ": disconnecting...");
            }
            else {
                debuglog('disconnecting...');
            }
            if (_this.client == null) {
                return Promise.resolve(null);
            }
            else {
                _this.disconnecting = true;
                _this.client.removeListener('disconnected', _this.reconnect);
                return _this.disconnectHelper()
                    .timeout(5000, 'failed to disconnect within 5 seconds, returning anyway')
                    .catch(Promise.TimeoutError, function (e) {
                    debuglog(e && e.message ? e.message : e);
                    return true;
                });
            }
        };
        this.disconnectHelper = function () {
            return new Promise(function (resolve, reject) {
                _this.client.once('disconnected', function () {
                    if (_this.client) {
                        _this.client.removeAllListeners();
                    }
                    _this.client = null;
                    _this.connected = false;
                    _this.disconnecting = false;
                    if (_this.key && _this.path) {
                        debuglog(_this.path + "/" + _this.key + ": disconnected");
                    }
                    else {
                        debuglog('disconnected');
                    }
                    resolve(true);
                });
                _this.client.close();
            });
        };
        /**
         * destroy the lock, disconnect and remove all listeners from the 'signal' event emitter
         * @returns {Promise<any>}
         */
        this.destroy = function () {
            return _this.disconnect().then(function () {
                if (_this.key && _this.path) {
                    debuglog(_this.path + "/" + _this.key + ": destroyed");
                }
                else {
                    debuglog("destroyed");
                }
                _this.removeAllListeners();
                // wait for session timeout for ephemeral lock to go away
                // return Promise.delay(this.config.sessionTimeout).thenReturn(true);
                return true;
            });
        };
        /**
         * internal method to reconnect, wired up to disconnect event of zk client
         * @returns {Promise<any>}
         */
        this.reconnect = function () {
            _this.connected = false;
            _this.retryCount++;
            if (_this.config.failImmediate || (_this.retryCount <= _this.config.retries) || !_this.locked) {
                return Promise.resolve(false);
            }
            debuglog('reconnecting...');
            return _this.connect(_this.config.spinDelay);
        };
        /**
         * unlock a lock, removing the key from zookeeper, and disconnecting
         * the zk client and all event listeners. By default this also destroys
         * the lock and removes event listeners on the locks 'signals' event
         * @param [destroy=true] - remove listeners from lock in addition
         * to disconnecting zk client on completion, defaults to true
         * @returns {Promise<any>}
         */
        this.unlock = function (destroy) {
            if (destroy === void 0) { destroy = true; }
            return new Promise(function (resolve, reject) {
                var cleanup = function () {
                    var destroyFunc;
                    if (destroy) {
                        destroyFunc = _this.destroy;
                    }
                    else {
                        destroyFunc = _this.disconnect;
                    }
                    destroyFunc().then(function () {
                        if (_this.path && _this.key) {
                            debuglog(_this.path + "/" + _this.key + ": unlocked, cleanup complete");
                        }
                        else {
                            debuglog('cleanup complete');
                        }
                        resolve(true);
                    }).catch(function () {
                        debuglog('cleanup failed');
                        reject(false);
                    });
                };
                if (_this.client) {
                    if (_this.path && _this.key) {
                        debuglog(_this.path + "/" + _this.key + ": unlocking..,");
                        _this.client.remove(_this.path + "/" + _this.key, function (err) {
                            if (err && err.message && err.message.indexOf('NO_NODE') < 0) {
                                debuglog(_this.path + "/" + _this.key + ": failed to remove due to: " + err.message + ".");
                            }
                            cleanup();
                        });
                    }
                    else {
                        debuglog("lock not set, skipping unlock, but cleaning up connection");
                        cleanup();
                    }
                }
                else {
                    debuglog("client not connected, skipping unlock and cleanup");
                    resolve(true);
                }
            });
        };
        /**
         * wait for a lock to become free for a given key and acquire it, with an optional
         * timeout upon which the lock will fail. if not currently connected to zookeeper,
         * this will connect, and on timeout, the lock will disconnect from zookeeper
         * @param key
         * @param [timeout]
         * @returns {Promise<any>}
         */
        this.lock = function (key, timeout) {
            if (timeout === void 0) { timeout = 0; }
            var path = "/locks/" + (_this.config.pathPrefix ? _this.config.pathPrefix + '/' : '');
            var nodePath = "" + path + key;
            var someRandomExtraLogText = _this.config.maxConcurrentHolders > 1 ?
                " with " + _this.config.maxConcurrentHolders + " concurrent lock holders" :
                '';
            if (_this.locked) {
                debuglog('already locked');
                return Promise.resolve(_this);
            }
            debuglog("try locking " + key + " at " + path + someRandomExtraLogText);
            if (timeout && !_this.config.failImmediate) {
                return _this.lockHelper(path, nodePath, timeout)
                    .timeout(timeout, 'timeout')
                    .catch(Promise.TimeoutError, function (te) {
                    _this.timedOut = true;
                    _this.disconnect();
                    throw new ZookeeperLockTimeoutError('timeout', path, timeout);
                });
            }
            else {
                return _this.lockHelper(path, nodePath);
            }
        };
        this.lockHelper = function (path, nodePath, timeout) {
            return _this.connect()
                .then(function () {
                _this.timedOut = false;
                debuglog("making lock at " + nodePath);
                return _this.makeLockDir(nodePath);
            })
                .then(function () {
                return _this.initLock(nodePath);
            })
                .then(function () {
                debuglog(_this.path + "/" + _this.key + ": waiting for lock");
                return _this.waitForLock(nodePath);
            })
                .then(function () {
                _this.locked = true;
                debuglog(_this.path + "/" + _this.key + ": lock acquired");
                return _this;
            }).catch(function (err) {
                _this.locked = false;
                debuglog(_this.path + "/" + _this.key + ": error grabbing lock: " + err.message);
                throw err;
            });
        };
        /**
         * make the zk node that will hold the locks if it doens't already exist
         * @param path
         * @returns {Promise<any>}
         */
        this.makeLockDir = function (path) {
            return new Promise(function (resolve, reject) {
                _this.client.mkdirp(path, function (err) {
                    if (_this.checkBail()) {
                        return;
                    }
                    if (err) {
                        reject(new Error("Failed to create directory: " + path + " due to: " + err + "."));
                        return;
                    }
                    resolve(true);
                });
            });
        };
        /**
         * create a lock as a ephemeral sequential child node of the supplied path, prefixed with 'lock-',
         * to state intent to acquire a lock
         * @param path
         * @returns {Promise<any>}
         */
        this.initLock = function (path) {
            return new Promise(function (resolve, reject) {
                _this.client.create(path + "/lock-", new Buffer('lock'), zk.CreateMode.EPHEMERAL_SEQUENTIAL, function (err, lockPath) {
                    if (_this.checkBail()) {
                        return;
                    }
                    if (err) {
                        reject(new Error("Failed to create node: " + lockPath + " due to: " + err + "."));
                        return;
                    }
                    debuglog("init lock: " + path + ", " + lockPath.replace(path + '/', ''));
                    _this.path = path;
                    _this.key = lockPath.replace(path + '/', '');
                    resolve(true);
                });
            });
        };
        /**
         * loop until lock is available or timeout occurs
         * @param path
         * @returns {Promise<any>}
         */
        this.waitForLock = function (path) {
            return new Promise(function (resolve, reject) {
                _this.waitForLockHelper(resolve, reject, path);
            });
        };
        this.checkBail = function () {
            return _this.timedOut || _this.disconnecting || !_this.connected;
        };
        /**
         * helper method that does the grunt of the work of waiting for the lock. This method does 2 things, first
         * reads the lock path to compare the locks key to the other keys that are children of the path. if this locks
         * sequence number is the lowest, the lock has been aquired. If not, this method reactively responds to
         * children changed events from the zk-client for the path we want to aqcuire the lock for, and recurses to
         * repeat this process until the sequence is the lowest
         * @param resolve
         * @param reject
         * @param path
         */
        this.waitForLockHelper = function (resolve, reject, path) {
            debuglog(path + " wait loop.");
            if (_this.locked || _this.checkBail()) {
                return;
            }
            _this.client.getChildren(path, function (event) {
                if (_this.locked || _this.checkBail()) {
                    return;
                }
                debuglog(path + "/" + _this.key + ": children changed.");
                _this.waitForLockHelper(resolve, reject, path);
            }, function (err, locks, state) {
                if (_this.locked || _this.checkBail()) {
                    return;
                }
                try {
                    if (err || !locks || locks.length === 0) {
                        var errMsg = err && err.message ? err.message : 'no children';
                        debuglog(path + "/" + _this.key + ": failed to get children: " + errMsg);
                        return reject(new Error("Failed to get children node: " + errMsg + "."));
                    }
                    var sequence = _this.filterLocks(locks)
                        .map(function (l) {
                        return ZookeeperLock.getSequenceNumber(l);
                    })
                        .filter(function (l) {
                        return l >= 0;
                    });
                    debuglog(path + "/" + _this.key + ": lock sequence: " + JSON.stringify(sequence));
                    var mySeq = ZookeeperLock.getSequenceNumber(_this.key);
                    // todo: fix me for max concurrent holders...
                    // make sure we are first...
                    var min = sequence.reduce(function (acc, elem) {
                        return Math.min(acc, elem);
                    }, mySeq);
                    debuglog(path + "/" + _this.key + ": checking " + mySeq + " less than " + min + " + " + _this.config.maxConcurrentHolders);
                    if (mySeq < (min + _this.config.maxConcurrentHolders)) {
                        debuglog(path + "/" + _this.key + ": " + mySeq + " can grab the lock on " + path);
                        return resolve(true);
                    }
                    else if (_this.config.failImmediate) {
                        _this.locked = false;
                        debuglog(path + "/" + _this.key + ": failing immediately");
                        return reject(new ZookeeperLockAlreadyLockedError('already locked', path));
                    }
                    _this.locked = false;
                    debuglog(path + "/" + _this.key + ": lock not available for " + mySeq + " on " + path + ", waiting...");
                }
                catch (ex) {
                    debuglog(path + "/" + _this.key + ": error - " + ex.message);
                    reject(ex);
                }
            });
        };
        /**
         * method to filter zk node children to contain only those that are prefixed with 'lock-',
         * which are assumed to be created by this library
         * @param children
         * @returns {string[]|T[]}
         */
        this.filterLocks = function (children) {
            var filtered = children.filter(function (l) {
                return l !== null && l.indexOf('lock-') === 0;
            });
            return filtered;
        };
        /**
         * check if a lock exists, connecting to zk client if not connected
         * @param key
         * @returns {Promise<boolean>}
         */
        this.checkLocked = function (key) {
            return _this.connect()
                .then(function () {
                return _this.checkedLockedHelper(key);
            })
                .catch(function (err) {
                if (err && err.message && err.message.indexOf('NO_NODE') > -1) {
                    return false;
                }
                else {
                    debuglog("error checking locked: " + key + ": " + (err && err.message ? err.message : 'unknown'));
                    throw err;
                }
            });
        };
        this.checkedLockedHelper = function (key) {
            return new Promise(function (resolve, reject) {
                var path = "/locks/" + (_this.config.pathPrefix ? _this.config.pathPrefix + '/' : '');
                var nodePath = "" + path + key;
                _this.client.getChildren(nodePath, null, function (err, locks, stat) {
                    if (err) {
                        reject(err);
                    }
                    else if (locks) {
                        var filtered = _this.filterLocks(locks);
                        debuglog("check " + nodePath + ": " + JSON.stringify(filtered));
                        if (filtered && (filtered.length - (_this.config.maxConcurrentHolders - 1) > 0)) {
                            debuglog("check " + nodePath + ": no locks held");
                            resolve(true);
                        }
                        else {
                            debuglog("check " + nodePath + ": " + filtered.length + " locks held");
                            resolve(false);
                        }
                    }
                    else {
                        resolve(false);
                    }
                });
            });
        };
        this.config = config ? config : {};
        if (this.config.sessionTimeout == null) {
            this.config.sessionTimeout = 15000;
        }
        if (this.config.spinDelay == null) {
            this.config.spinDelay = 0;
        }
        if (this.config.retries == null) {
            this.config.retries = 0;
        }
        if (this.config.maxConcurrentHolders == null) {
            this.config.maxConcurrentHolders = 1;
        }
        debuglog(JSON.stringify(this.config));
    }
    /**
     * create a zookeeper client which powers the lock, done when creating a new lock
     * or zk connection expires
     * @returns {Promise<any>}
     */
    ZookeeperLock.prototype.createClient = function () {
        var _this = this;
        debuglog('creating client');
        if (this.client != null) {
            debuglog('client already created');
            return Promise.resolve(true);
        }
        else {
            return this.config.serverLocator().then(function (location) {
                var server = location.host;
                if (location.port) {
                    server += ":" + location.port;
                }
                debuglog('server location resolved');
                debuglog(server);
                var client = zk.createClient(server, {
                    sessionTimeout: _this.config.sessionTimeout,
                    spinDelay: _this.config.spinDelay,
                    retries: _this.config.retries
                });
                _this.client = client;
                _this.client.once('expired', function () {
                    if (_this.key && _this.path) {
                        debuglog(_this.path + "/" + _this.key + ": expired");
                    }
                    else {
                        debuglog('expired');
                    }
                    _this.emit(ZookeeperLock.Signals.LOST);
                    if (_this.client) {
                        debuglog('removing listeners');
                        _this.client.removeAllListeners();
                        _this.removeAllListeners();
                        _this.client.close();
                        _this.connected = false;
                        _this.client = null;
                    }
                });
                _this.client.once('disconnected', _this.reconnect);
                return true;
            });
        }
    };
    ;
    ZookeeperLock.config = null;
    ZookeeperLock.Signals = {
        LOST: 'lost',
        TIMEOUT: 'timeout'
    };
    /**
     * set static config to use by static helper methods
     * @param config
     */
    ZookeeperLock.initialize = function (config) {
        ZookeeperLock.config = config;
    };
    /**
     * create a new lock using the static stored config
     * @returns {ZookeeperLock}
     */
    ZookeeperLock.lockFactory = function () {
        return new ZookeeperLock(ZookeeperLock.config);
    };
    /**
     * create a new lock and lock it using the static stored config, with optional timeout
     * @param key
     * @param timeout
     * @returns {Promise<ZookeeperLock>}
     */
    ZookeeperLock.lock = function (key, timeout) {
        var zkLock = new ZookeeperLock(ZookeeperLock.config);
        return zkLock.lock(key, timeout).catch(function (err) {
            zkLock.destroy();
            throw err;
        });
    };
    /**
     * check if a lock exists for a path using the static config
     * @param key
     * @returns {Promise<boolean>}
     */
    ZookeeperLock.checkLock = function (key) {
        var zkLock = new ZookeeperLock(ZookeeperLock.config);
        return zkLock.checkLocked(key)
            .then(function (result) {
            return result ? true : false;
        })
            .finally(function () {
            zkLock.destroy();
        });
    };
    /**
     * get the numeric part of the lock key
     * @param path
     * @returns {number}
     */
    ZookeeperLock.getSequenceNumber = function (path) {
        return parseInt(path.replace('lock-', ''), 10);
    };
    return ZookeeperLock;
}(events_1.EventEmitter));
exports.ZookeeperLock = ZookeeperLock;
//# sourceMappingURL=zookeeperLock.js.map